/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 1997-2000 Tempest s.r.o. http://www.tempest.sk
 * Copyright (C) 2001-2002 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

#include <autoconf.h>

#include <sys/types.h>	/* size_t */
#include <sys/stat.h>	/* stat() */
#include <errno.h>		/* errno */
#include <fcntl.h>		/* open() */
#include <stdio.h>		/* sprintf() */
#include <string.h>		/* strchr() */
#include <unistd.h>		/* stat() */

#include "clex.h"
#include "cfg.h"

#include "control.h"	/* control_loop() */
#include "directory.h"	/* filepos_set() */
#include "edit.h"		/* edit_putstr() */
#include "err.h"		/* err_exit() */
#include "exec.h"		/* exec_initialize() */
#include "inout.h"		/* win_remark() */
#include "list.h"		/* list_initialize() */
#include "sort.h"		/* sort_files() */
#include "start.h"		/* restart_clex() */
#include "userdata.h"	/* get_myhomedir() */
#include "ustring.h"	/* USTRING() */
#include "util.h"		/* estrdup() */
#include "xterm_title.h"/* xterm_title_set() */

#define MAX_CFG_LINES 100	/* config file size limit */

extern int errno;

static char *user_config_file;

typedef struct {
	int code;
	int min, max, initial, current, new;				/* values */
	char *desc;		/* enumerated: show this text instead of numbers */
} CNUM;

typedef struct {
	int code;
	char *initial;
	char current[CFGVALUE_MAXLEN + 1], new[CFGVALUE_MAXLEN + 1];
} CSTR;

static CNUM table_numeric[] = {
	/* enumerated */
	{ CFG_CMD_LINES,	 2,   4,  2, 0, 0, "2\0003\0004" },
	{ CFG_FRAME,		 0,   2,  0, 0, 0, "---\0===\0line graphics" },
	{ CFG_GROUP_FILES,	 0,   2,  2, 0, 0, "Do not group\0"
		"Group: directories, special files, plain files\0"
		"Group: directories, devices, special files, plain files" },
	{ CFG_SHOW_HIDDEN,	 0,   1,  1, 0, 0, "Do not show\0Show" },
	{ CFG_WARN_CD,		 0,   1,  1, 0, 0, "No\0Yes" },
	{ CFG_WARN_RM,		 0,   1,  1, 0, 0, "No\0Yes" },
	{ CFG_WARN_LONG,	 0,   1,  1, 0, 0, "No\0Yes" },
	{ CFG_WARN_SELECT,	 0,   1,  1, 0, 0, "No\0Yes" },
	{ CFG_XTERM_TITLE,	 0,   2,  1, 0, 0, "No\0Auto\0Yes" },
	/* really numeric */
	{ CFG_COMPLETIONS,	10, 100, 30, 0, 0, 0 },
	{ CFG_DIRECTORIES,	10, 100, 60, 0, 0, 0 },
	{ CFG_HISTORY,		10, 100, 40, 0, 0, 0 }
};

static CSTR table_string[] = {
	{ CFG_CMD_F3,		"more $f",		"", "" },
	{ CFG_CMD_F4,		"vi $f",		"", "" },
	{ CFG_CMD_F5,		"cp -ir $f $2",	"", "" },
	{ CFG_CMD_F6,		"mv -i $f $2",	"", "" },
	{ CFG_CMD_F7,		"mkdir ",		"", "" },
	{ CFG_CMD_F8,		"rm $f",		"", "" },
	{ CFG_CMD_F9,		"lpr $f",		"", "" },
	{ CFG_CMD_F10,		"",				"", "" },
	{ CFG_CMD_F11,		"",				"", "" },
	{ CFG_CMD_F12,		"",				"", "" },
	{ CFG_DIR2,			"",				"", "" },
	{ CFG_LAYOUT,		"dst:mpoA",		"", "" },
	{ CFG_LOCALE,		"",				"", "" },
	{ CFG_HELPFILE,		HELP_FILE,		"", "" },
	{ CFG_PROMPT, 		"",				"", "" },
	{ CFG_QUOTE,		"",				"", "" },
	{ CFG_SHELLPROG,	"",				"", "" }
};

/*
 * help - everything in one place and sorted (for easy editing),
 * must fit on minimal width screen (MIN_COLS)
 */
static struct {
	int code;
	char *help;
} table_help[] = {
	{ CFG_CMD_F3,		"Command F3 = view file(s)" },
	{ CFG_CMD_F4,		"Command F4 = edit file(s)" },
	{ CFG_CMD_F5,		"Command F5 = copy file(s)" },
	{ CFG_CMD_F6,		"Command F6 = move file(s)" },
	{ CFG_CMD_F7,		"Command F7 = make directory" },
	{ CFG_CMD_F8,		"Command F8 = remove file(s)" },
	{ CFG_CMD_F9,		"Command F9 = print file(s)" },
	{ CFG_CMD_F10,		"Command F10 = user defined" },
	{ CFG_CMD_F11,		"Command F11 = user defined" },
	{ CFG_CMD_F12,		"Command F12 = user defined" },
	{ CFG_CMD_LINES,	"How many lines takes the command "
						"line on the screen" },
	{ CFG_COMPLETIONS,	"Completions panel size "
						"(restart after change)" },
	{ CFG_DIR2,			"Initial working directory for the secondary "
						"file panel" },
	{ CFG_DIRECTORIES,	"Directory panel size (restart after change)" },
	{ CFG_FRAME,		"Frame style: plain '---' or '===' "
						"or special line graphics" },
	{ CFG_GROUP_FILES,	"Group files of the same type together" },
	{ CFG_HELPFILE,		"Where the on-line help is stored "
						"(restart after change)" },
	{ CFG_HISTORY,		"History panel size (restart after change)" },
	{ CFG_LAYOUT,		"File panel layout, see help" },
	{ CFG_LOCALE,		"Local language rules "
						"(leave blank for autodetect)" },
	{ CFG_PROMPT,		"Command line prompt (leave blank "
						"for autoselect)" },
	{ CFG_QUOTE,		"Additional filename chars to be quoted, "
						"see help" },
	{ CFG_SHELLPROG,	"Shell program, see help (leave blank "
						"for autodetect)" },
	{ CFG_SHOW_HIDDEN,	"Show hidden files (their filenames begin "
						"with a dot)" },
	{ CFG_WARN_CD,		"Warn that 'cd' (change directory) command "
						"will not work" },
	{ CFG_WARN_RM,		"Warn before executing 'rm' (remove) command" },
	{ CFG_WARN_LONG,	"Warn that the command line is too long to be "
						"displayed" },
	{ CFG_WARN_SELECT,	"Remind that you have selected some files" },
	{ CFG_XTERM_TITLE,	"Change the X terminal window title" },
};

static CONFIG_ENTRY config[] = {
	{ "FRAME",			{ 0,0,0 }, 0, 0 },
	{ "GROUP_FILES",	{ 0,0,0 }, 0, 0 },
	{ "SHOW_HIDDEN",	{ 0,0,0 }, 0, 0 },
	{ "WARN_CD",		{ 0,0,0 }, 0, 0 },
	{ "WARN_RM",		{ 0,0,0 }, 0, 0 },
	{ "WARN_LONG",		{ 0,0,0 }, 0, 0 },
	{ "WARN_SELECT",	{ 0,0,0 }, 0, 0 },
	{ "XTERM_TITLE",	{ 0,0,0 }, 0, 0 },
	{ "CMD_LINES",		{ 0,0,0 }, 0, 0 },
	{ "CMD_F3",			{ 0,0,0 }, 0, 0 },
	{ "CMD_F4",			{ 0,0,0 }, 0, 0 },
	{ "CMD_F5",			{ 0,0,0 }, 0, 0 },
	{ "CMD_F6",			{ 0,0,0 }, 0, 0 },
	{ "CMD_F7",			{ 0,0,0 }, 0, 0 },
	{ "CMD_F8",			{ 0,0,0 }, 0, 0 },
	{ "CMD_F9",			{ 0,0,0 }, 0, 0 },
	{ "CMD_F10",		{ 0,0,0 }, 0, 0 },
	{ "CMD_F11",		{ 0,0,0 }, 0, 0 },
	{ "CMD_F12",		{ 0,0,0 }, 0, 0 },
	{ "DIR2",			{ 0,0,0 }, 0, 0 },
	{ "LOCALE",			{ 0,0,0 }, 0, 0 },
	{ "HELPFILE",		{ 0,0,0 }, 0, 0 },
	{ "QUOTE",			{ 0,0,0 }, 0, 0 },
	{ "LAYOUT",			{ 0,0,0 }, 0, 0 },
	{ "PROMPT",			{ 0,0,0 }, 0, 0 },
	{ "SHELLPROG",		{ 0,0,0 }, 0, 0 },
	{ "HISTORY",		{ 0,0,0 }, 0, 0 },
	{ "DIRECTORIES",	{ 0,0,0 }, 0, 0 },
	{ "COMPLETIONS",	{ 0,0,0 }, 0, 0 }
};	/* must exactly match CFG_XXX #defines */

/* 'move' values MOV_X2Y understood by set_value() */
#define MOV_SRC_CURRENT	 1
#define MOV_SRC_INITIAL	 2
#define MOV_SRC_NEW		 4
#define MOV_DST_CURRENT	 8
#define MOV_DST_NEW		16
#define MOV_N2C	(MOV_SRC_NEW     | MOV_DST_CURRENT)
#define MOV_C2N	(MOV_SRC_CURRENT | MOV_DST_NEW)
#define MOV_I2N	(MOV_SRC_INITIAL | MOV_DST_NEW)
#define MOV_I2C	(MOV_SRC_INITIAL | MOV_DST_CURRENT)

static void
set_value(int code, int move)
{
	int src_num, *dst_num;
	char *src_str, *dst_str;
	CNUM *pnum;
	CSTR *pstr;

	if (config[code].flag.isnum) {
		pnum = config[code].table;
		if (move & MOV_SRC_CURRENT)
			src_num = pnum->current;
		else if (move & MOV_SRC_INITIAL)
			src_num = pnum->initial;
		else 
			src_num = pnum->new;
		if (move & MOV_DST_CURRENT)
			dst_num = &pnum->current;
		else
			dst_num = &pnum->new;
		*dst_num = src_num;
	}
	else {
		pstr = config[code].table;
		if (move & MOV_SRC_CURRENT)
			src_str = pstr->current;
		else if (move & MOV_SRC_INITIAL)
			src_str = pstr->initial;
		else
			src_str = pstr->new;
		if (move & MOV_DST_CURRENT)
			dst_str = pstr->current;
		else
			dst_str = pstr->new;
		strcpy(dst_str,src_str);
	}
}

static CONFIG_ENTRY *
get_variable(const char *var)
{
	size_t i;

	for (i = 0; i < ARRAY_SIZE(config); i++)
		if (strcmp(config[i].var,var) == 0)
			return config + i;
	return 0;
}

static CNUM *
get_numeric(int code)
{
	size_t i;

	for (i = 0; i < ARRAY_SIZE(table_numeric); i++)
		if (table_numeric[i].code == code)
			return table_numeric + i;
	return 0;
}

static CSTR *
get_string(int code)
{
	size_t i;

	for (i = 0; i < ARRAY_SIZE(table_string); i++)
		if (table_string[i].code == code)
			return table_string + i;
	return 0;
}

static const char *
get_help(int code)
{
	size_t i;

	for (i = 0; i < ARRAY_SIZE(table_help); i++)
		if (table_help[i].code == code)
			return table_help[i].help;
	return 0;
}

static void
config_error(const char *file, int line, const char *errmsg)
{
	static const char *last_file = 0;

	if (file != last_file)
		printf("CONFIG: Reading configuration file \"%s\"\n",
		  last_file = file);
	fputs("CONFIG: ",stdout);
	if (line > 0)
		printf("Line %d: ",line);
	puts(errmsg);
	startup_warning();
}

static void
read_configuration(const char *file, int setdefaults)
{
	FILE *fp;
	char buff[CFGVAR_MAXLEN + CFGVALUE_MAXLEN + 16], *value;
	int ilen, num, lnum, len;
	struct stat fst;
	CONFIG_ENTRY *pce;
	CNUM *pnum;
	CSTR *pstr;

	/* few basic checks */
	if (stat(file,&fst) < 0 ) {
		/* it is OK if the file is missing */
		if (errno != ENOENT)
			config_error(file,0,"File is unreadable");
		return;
	}
	if ((fst.st_mode & S_IWOTH) == S_IWOTH) {
		/* little bit of security */
		config_error(file,0,"File is unsafe (world-writable), skipped");
		return;
	}
	if (!S_ISREG(fst.st_mode)) {
		/* sanity check */
		config_error(file,0,"This is not a plain file");
		return;
	}

	/*
	 * yes, there is a race condition here, but the checks
	 * above are merely detecting stupid mistakes
	 */

	fp = fopen(file,"r");
	if (fp == 0) {
		config_error(file,0,"File is unreadable");
		return;	
	}
	for (lnum = 1; fgets(buff,sizeof(buff),fp); lnum++) {
		if (lnum > MAX_CFG_LINES) {
			config_error(file,0,
			  "File is too long, skipping the rest of this file");
			break;
		}

		/* ignore comments */
		if (buff[0] == '\n' || buff[0] == '#')
			continue;

		/* strip newline */
		len = strlen(buff) - 1;
		if (buff[len] != '\n') {
			config_error(file,lnum,
			  "line is too long, skipping the rest of this file");
			break;	/* why bother to recover ? */
		}
		buff[len] = '\0';

		/* split VARIABLE and VALUE */
		if ( (value = strchr(buff,'=')) == 0) {
			config_error(file,lnum,
			  "incorrect syntax, format: VARIABLE=VALUE");
			continue;
		}
		*value++ = '\0';
		len -= value - buff;

		pce = get_variable(buff);
		if (pce == 0) {
			config_error(file,lnum,"unknown variable");
			continue;
		}

		if (pce->flag.isnum) {
			pnum = pce->table;
			if (sscanf(value," %d %n",&num,&ilen) < 1 || ilen != len)
				config_error(file,lnum,"invalid number");
			else if (num < pnum->min || num > pnum->max)
				config_error(file,lnum,"numeric value out of range");
			else if (setdefaults)
				pnum->initial = num;
			else
				pnum->current = num;
		} else {
			pstr = pce->table;
			if (len > CFGVALUE_MAXLEN)
				config_error(file,lnum,"string value is too long");
			else if (setdefaults)
				pstr->initial = estrdup(value);
			else
				strcpy(pstr->current,value);
		}
	}
	if (ferror(fp))
		config_error(file,0,"File read error");
	fclose(fp);
}

void
config_initialize(void)
{
	int fd;
	size_t i;

	static const char *menu[] = {
		"CANCEL:  no configuration change",
		"ACCEPT:  use the new configuration",
		"SAVE:    accept and save the new configuration "
		  "(file ~/.clexrc)",
		"RESTART: save the new configuration and restart CLEX"
	};
	static const char *admin_menu[] = {
		"QUIT:  no configuration change",
		"SAVE:  save the new system-wide defaults to " CONFIG_FILE
	};

	panel_config_confirm.type = PANEL_TYPE_MENU;
	panel_config_confirm.min = 0;
	if (admin) {
		panel_config_confirm.menu = admin_menu;
		panel_config_confirm.cnt = ARRAY_SIZE(admin_menu);
	}
	else {
		panel_config_confirm.menu = menu;
		panel_config_confirm.cnt = ARRAY_SIZE(menu);
	}

	panel_config.type = PANEL_TYPE_CONFIG;
	panel_config.min = -1;
	panel_config.cnt = ARRAY_SIZE(config);
	panel_config.config = config;

	/* initialize 'config' table */
	for (i = 0; i < ARRAY_SIZE(config); i++) {
		if ( (config[i].table = get_numeric(i)) )
			config[i].flag.isnum = 1;
		else if ( (config[i].table = get_string(i)) )
			config[i].flag.isnum = 0;
		else
			err_exit("BUG: config variable not defined (code %d)",i);
		if ( (config[i].help = get_help(i)) == 0)
			err_exit("BUG: no help for config variable (code %d)",i);
	}

	if (!admin) {
		/* user's configuration file name */
		pathname_set_directory(get_myhomedir());
		user_config_file = estrdup(pathname_join(".clexrc"));

		/* read defaults, initialize, read config */
		read_configuration(CONFIG_FILE,1);
		for (i = 0; i < ARRAY_SIZE(config); i++)
			set_value(i,MOV_I2C);
		read_configuration(user_config_file,0);
	}
	else {
		/* check user's access rights */
		fd = open(CONFIG_FILE,O_WRONLY);
		if (fd < 0) {
			if (errno == EACCES)
				err_exit("Only system administrators may update the "
				  "system-wide configuration.\n"
				  "      You don't have the required privilege.");
		}
		else
			close(fd);

		/* initialize, read defaults as config */
		for (i = 0; i < ARRAY_SIZE(config); i++)
			set_value(i,MOV_I2C);
		read_configuration(CONFIG_FILE,0);
	}
}

int
config_num(int code)
{
	if (!config[code].flag.isnum)
		err_exit("BUG: config variable type mismatch (code %d)",code);
	return ((CNUM *)config[code].table)->current;
}

const char *
config_str(int code)
{
	if (config[code].flag.isnum)
		err_exit("BUG: config variable type mismatch (code %d)",code);
	return ((CSTR *)config[code].table)->current;
}

const char *
config_print_value(int i)
{
	int val;
	char *desc;
	static char buff[16];
	CNUM *pnum;

	if (!config[i].flag.isnum)
		return ((CSTR *)(config[i].table))->new;

	pnum = config[i].table;
	val = pnum->new;
	desc = pnum->desc;
	if (desc) {
		while (val-- > pnum->min)
			while (*desc++)
				;
		return desc;
	}

	sprintf(buff,"%d",val);
	return buff;
}

static void
config_save(void)
{
	size_t i;
	int errflag;
	mode_t um;
	FILE *fp;

	if (admin) {
		umask(022);
		fp = fopen(CONFIG_FILE,"w");
	}
	else {
		um = umask(0);
		umask(um | 022);
		fp = fopen(user_config_file,"w");
		umask(um);
	}

	if (fp == 0) {
		win_warning("CONFIG: Cannot open configuration file "
		  "for writing.");
		return;
	}
	fprintf(fp,	"#\n"
				"# CLEX configuration file\n"
				"#\n");
	for (i = 0; i < ARRAY_SIZE(config); i++)
		if (config[i].flag.saveit) {
			if (config[i].flag.isnum)
				fprintf(fp,"%s=%d\n",config[i].var,
				  ((CNUM *)config[i].table)->new);
			else
				fprintf(fp,"%s=%s\n",config[i].var,
				  ((CSTR *)config[i].table)->new);
		}
	errflag = ferror(fp);
	if (fclose(fp) || errflag)
		win_warning("CONFIG: File write error occurred.");
}

void
config_prepare(void)
{
	size_t i;

	for (i = 0; i < ARRAY_SIZE(config); i++)
		set_value(i,MOV_C2N);
	panel_config.top = panel_config.curs = -1;
	panel = (PANEL *)&panel_config;
	textline = 0;
}

void
config_confirm_prepare(void)
{
	size_t i;
	int changed;
	CNUM *pnum;
	CSTR *pstr;

	for (changed = 0, i = 0; i < ARRAY_SIZE(config); i++) {
		if (config[i].flag.isnum) {
			pnum = config[i].table;
			config[i].flag.changed = pnum->new != pnum->current;
			config[i].flag.saveit = pnum->new != pnum->initial;
		}
		else {
			pstr = config[i].table;
			config[i].flag.changed =
			  strcmp(pstr->new,pstr->current) != 0;
			config[i].flag.saveit =
			  strcmp(pstr->new,pstr->initial) != 0;
		}
		if (config[i].flag.changed)
			changed = 1;
	}

	panel_config_confirm.top = 0;
	panel = (PANEL *)&panel_config_confirm;
	textline = 0;
	if (!changed) {
		win_remark("configuration was not changed");
		panel_config_confirm.curs = 0;	/* CANCEL line */
	}
	else
		panel_config_confirm.curs = 1;	/* ACCEPT line */
}

void
config_edit_num_prepare(void)
{
	static char prompt[CFGVAR_MAXLEN + 48];
	char line[16];
	CNUM *pnum;

	pnum = config[panel_config.curs].table;
	
	textline = &line_tmp;
	sprintf(prompt,"%s (range: %d - %d): ",
	  config[panel_config.curs].var,pnum->min,pnum->max);
	edit_setprompt(textline,prompt);

	sprintf(line,"%d",pnum->new);
	edit_nu_putstr(line);
	edit_adjust();
}

void
config_edit_str_prepare(void)
{
	static char prompt[CFGVAR_MAXLEN + 32];

	textline = &line_tmp;
	sprintf(prompt,"%s (%d chars max): ",
	  config[panel_config.curs].var,CFGVALUE_MAXLEN);
	edit_setprompt(textline,prompt);

	edit_nu_putstr(((CSTR *)config[panel_config.curs].table)->new);
	edit_adjust();
}

void
cx_config_edit(void)
{
	CNUM *pnum;

	if (panel_config.curs < 0) {
		next_mode = MODE_CONFIG_CONFIRM;
		return;
	}

	if (config[panel_config.curs].flag.isnum) {
		pnum = config[panel_config.curs].table;
		if (pnum->desc) {
			if (++pnum->new > pnum->max)
				pnum->new = pnum->min;
		}
		else
			control_loop(MODE_CONFIG_EDIT_NUM);
	}
	else
		control_loop(MODE_CONFIG_EDIT_STR);
	win_panel_opt();
}

void
cx_config_num_enter(void)
{
	int num, ilen;
	CNUM *pnum;

	pnum = config[panel_config.curs].table;
	if (sscanf(USTR(textline->line)," %d %n",&num,&ilen) < 1 ||
	  ilen != textline->size)
		win_remark("numeric value required");
	else if (num < pnum->min || num > pnum->max)
		win_remark("value is out of range");
	else {
		pnum->new = num;
		next_mode = MODE_SPECIAL_RETURN;
	}
}

void
cx_config_str_enter(void)
{
	if (textline->size > CFGVALUE_MAXLEN)
		win_remark("string is too long");
	else {
		strcpy(((CSTR *)config[panel_config.curs].table)->new,
		  USTR(textline->line));
		next_mode = MODE_SPECIAL_RETURN;
	}
}

void
cx_config_default(void)
{
	if (panel_config.curs < 0)
		return;
	set_value(panel_config.curs,MOV_I2N);
	win_panel_opt();
}

void
cx_config_original(void)
{
	if (panel_config.curs < 0)
		return;
	set_value(panel_config.curs,MOV_C2N);
	win_panel_opt();
}

static void
config_apply(void)
{
	size_t i;
	int reread, sort, prompt;

	for (i = 0; i < ARRAY_SIZE(config); i++)
		if (config[i].flag.changed)
			set_value(i,MOV_N2C);

	reread = sort = prompt = 0;
	if (config[CFG_FRAME].flag.changed)
		win_frame();
	if (config[CFG_GROUP_FILES].flag.changed)
		sort = 1;
	if (config[CFG_SHOW_HIDDEN].flag.changed)
		reread = 1;
	if (config[CFG_XTERM_TITLE].flag.changed) {
		xterm_title_restore();
		xterm_title_initialize();
		xterm_title_set();
	}
	if (config[CFG_CMD_LINES].flag.changed)
		curses_draw_all();
	if (config[CFG_LOCALE].flag.changed)
		reread = 1;
	if (config[CFG_LAYOUT].flag.changed) {
		win_set_layout();
		reread = 1;
	}
	if (config[CFG_PROMPT].flag.changed)
		prompt = 1;
	if (config[CFG_SHELLPROG].flag.changed) {
		exec_initialize();
		prompt = 1;
	}

	if (prompt)
		exec_setprompt_cmd();
	if (sort && !reread) {
		/* sort the primary panel */
		filepos_save();
		sort_files();
		filepos_set();
		/* sort the secondary panel */
		ppanel_file = ppanel_file->other;
		filepos_save();
		sort_files();
		filepos_set();
		ppanel_file = ppanel_file->other;
	}
	else if (reread) {
		ppanel_file->other->expired = 1;
		list_initialize();
		list_directory();
	}
}

void
cx_config_confirm(void)
{
	if (admin) {
		if (panel_config_confirm.curs == 1)
			config_save();
		next_mode = MODE_SPECIAL_QUIT;
		return;
	}

	switch (panel_config_confirm.curs) {
	case 1: /* ACCEPT */
		config_apply();
		break;
	case 2: /* SAVE */
		config_apply();
		config_save();
		break;
	case 3: /* RESTART */
		/* config_save() must work even without config_apply() */
		config_save();
		restart_clex();
		/* NOTREACHED */
	}
	next_mode = MODE_SPECIAL_RETURN;
}
