/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 1997-2000 Tempest s.r.o. http://www.tempest.sk
 * Copyright (C) 2001-2002 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

#include <autoconf.h>

#include <sys/types.h>		/* clex.h */
#include <ctype.h>			/* iscntrl() */
#ifdef HAVE_NCURSES_H
# include <ncurses.h>		/* KEY_xxx */
#else
# include <curses.h>
#endif

#include "clex.h"
#include "control.h"

#include "cfg.h"			/* config_prepare() */
#include "completion.h"		/* compl_prepare() */
#include "directory.h"		/* dir_main_prepare() */
#include "edit.h"			/* cx_edit_xxx() */
#include "err.h"			/* err_exit() */
#include "filepanel.h"		/* cx_files_xxx() */
#include "help.h"			/* help_prepare() */
#include "history.h"		/* history_prepare() */
#include "inout.h"			/* win_panel() */
#include "locate.h"			/* locate_prepare() */
#include "panel.h"			/* cx_pan_xxx() */
#include "select.h"			/* select_prepare() */
#include "sort.h"			/* sort_prepare() */

/*
 * This is the main control section. It is table driven.
 *
 * PANEL is the main part of the screen, it shows various data
 *   depending on the panel type, the user can scroll through it.
 *
 * TEXTLINE is a line of text where the user can enter and edit
 *   his/her input.
 *
 * KEY_BINDING contains a keystroke and a corresponding function
 *   to be called every time that key is pressed. All such handler
 *   function names begin with the cx_ prefix.
 *
 * CLEX operation mode is defined by a PANEL, TEXTLINE, and a set of
 *   KEY_BINDINGs tables. The PANEL and TEXTLINE are initialized by
 *   a so-called preparation function after each mode change.
 *
 * Operation mode can be changed in one of two ways:
 *   - straightforward transition from mode A to mode B; this
 *     is achieved by simply setting the 'next_mode' global variable
 *   - nesting of modes; this is achieved by calling another
 *     instance of 'control_loop()'. To return back the variable
 *     'next_mode' must be set to MODE_SPECIAL_RETURN.
 */

typedef struct {
	int escp;			/* flag: press escape key first */
	int key;			/* if this key was pressed ... */
	void (*fn)(void);	/* ... then this function is to be invoked */
} KEY_BINDING;

#define CXM(X,M) void cx_mode_ ## X (void) { control_loop(MODE_ ## M); }
static CXM(config_main,CONFIG_MAIN)
static CXM(dir_main,DIR_MAIN)
static CXM(history,HIST)
static CXM(help,HELP)
static CXM(locate,LOCATE)
static CXM(menu,MENU)
static CXM(select,SELECT)
static CXM(sort,SORT)
static CXM(deselect,DESELECT)

#define CXT(X,M) void cx_trans_ ## X (void) { next_mode = MODE_ ## M; }
static CXT(config_confirm,CONFIG_CONFIRM)
static CXT(quit,SPECIAL_QUIT)
static CXT(return,SPECIAL_RETURN)

static KEY_BINDING tab_common[] = {
	{ 0,  CH_CTRL('C'),	cx_trans_return	},
	{ 1,  'q',			cx_trans_quit	},
	{ 1,  'x',			cx_trans_quit	},
	{ 0,  CH_CTRL('G'),	cx_menu_na		},	/* catch bogus ctrl-G */
	{ 0,  KEY_F(1),		cx_mode_help	},
	{ 0,  0,			0				}
};

static KEY_BINDING tab_compl[] = {
	{ 0,  CH_CTRL('I'),	cx_compl_complete	},
	{ 0,  CH_CTRL('M'),	cx_compl_complete	},
	{ 0,  0,			0					}
};

static KEY_BINDING tab_config[] = {
	{ 0,  CH_CTRL('C'),	cx_trans_config_confirm	},
	{ 0,  CH_CTRL('M'),	cx_config_edit			},
	{ 0,  'd',			cx_config_default		},
	{ 0,  'u',			cx_config_original		},
	{ 0,  0,			0						}
};

static KEY_BINDING tab_config_confirm[] = {
	{ 0,  CH_CTRL('M'),	cx_config_confirm	},
	{ 0,  0,			0					}
};

static KEY_BINDING tab_config_edit_num[] = {
	{ 0,  CH_CTRL('M'),	cx_config_num_enter	},
	{ 0,  0,			0					}
};

static KEY_BINDING tab_config_edit_str[] = {
	{ 0,  CH_CTRL('M'),	cx_config_str_enter	},
	{ 0,  0,			0					}
};

static KEY_BINDING tab_dir[] = {
	{ 0,  CH_CTRL('I'),	cx_dir_tab			},
	{ 0,  CH_CTRL('M'),	cx_dir_enter		},
	{ 0,  0,			0					}
};

static KEY_BINDING tab_edit[] = {
	{ 0,  CH_CTRL('B'),	cx_edit_w_left		},
	{ 0,  CH_CTRL('D'),	cx_edit_w_del		},
	{ 0,  CH_CTRL('H'),	cx_edit_backsp		},
	{ 0,  KEY_BACKSPACE,cx_edit_backsp		},
	{ 0,  CH_CTRL('U'),	cx_edit_kill		},
	{ 0,  CH_CTRL('V'),	cx_edit_insert_spc	},
	{ 0,  CH_CTRL('W'),	cx_edit_w_right		},
	{ 0,  CH_CTRL('Y'),	cx_edit_delend		},
	{ 0,  '\177',		cx_edit_delchar		},
	{ 0,  KEY_DC,		cx_edit_delchar		},
	{ 0,  KEY_LEFT,		cx_edit_left		},
	{ 0,  KEY_RIGHT,	cx_edit_right		},
	{ 0,  KEY_HOME,		cx_edit_begin		},
#ifdef KEY_END
	{ 0,  KEY_END,		cx_edit_end			},
#endif
	{ 1,  KEY_UP,		cx_edit_up			},
	{ 1,  KEY_DOWN,		cx_edit_down		},
	{ 0,  0,			0					}
};

static KEY_BINDING tab_editcmd[] = {
	{ 0,  CH_CTRL('E'),	cx_edit_paste_dir	},
	{ 0,  CH_CTRL('I'),	cx_files_tab		},
	{ 1,  CH_CTRL('I'),	cx_edit_fullpath	},
	{ 0,  CH_CTRL('M'),	cx_files_enter		},
	{ 1,  CH_CTRL('M'),	cx_files_cd			},
	{ 0,  CH_CTRL('K'),	cx_hist_complete	},
	{ 0,  CH_CTRL('N'),	cx_hist_next		},
	{ 0,  CH_CTRL('G'),	cx_mode_menu		},
	{ 0,  CH_CTRL('P'),	cx_hist_prev		},
	{ 0,  CH_CTRL('R'),	cx_files_reread		},
	{ 0,  CH_CTRL('T'),	cx_select_toggle	},
	{ 0,  CH_CTRL('X'),	cx_files_exchange	},
	{ 0,  KEY_IC,		cx_select_toggle	},
	{ 0,  KEY_IL,		cx_select_toggle	},
	{ 0,  KEY_F(2),		cx_edit_cmd_f2		},
	{ 0,  KEY_F(3),		cx_edit_cmd_f3		},
	{ 0,  KEY_F(4),		cx_edit_cmd_f4		},
	{ 0,  KEY_F(5),		cx_edit_cmd_f5		},
	{ 0,  KEY_F(6),		cx_edit_cmd_f6		},
	{ 0,  KEY_F(7),		cx_edit_cmd_f7		},
	{ 0,  KEY_F(8),		cx_edit_cmd_f8		},
	{ 0,  KEY_F(9),		cx_edit_cmd_f9		},
	{ 0,  KEY_F(10),	cx_edit_cmd_f10		},
	{ 0,  KEY_F(11),	cx_edit_cmd_f11		},
	{ 0,  KEY_F(12),	cx_edit_cmd_f12		},
	{ 0,  0,			0					}
};

static KEY_BINDING tab_help[] = {
	{ 0,  CH_CTRL('H'),	cx_help_back		},
	{ 0,  KEY_BACKSPACE,cx_help_back		},
	{ 0,  CH_CTRL('M'),	cx_help_link		},
	{ 0,  KEY_LEFT,		cx_help_back		},
	{ 0,  KEY_RIGHT,	cx_help_link		},
	{ 0,  KEY_F(1),		cx_help_contents	},	/* avoid F1 recursion */
	{ 0,  0,			0					}
};

static KEY_BINDING tab_hist[] = {
	{ 0,  CH_CTRL('I'),	cx_hist_paste	},
	{ 0,  CH_CTRL('M'),	cx_hist_enter	},
	{ 0,  CH_CTRL('N'),	cx_pan_up		},	/* redefine history next */
	{ 0,  CH_CTRL('P'),	cx_pan_down		},	/* redefine history prev */
	{ 0,  0,			0				}
};

/* pseudo-table returned by action() */
static KEY_BINDING tab_insertchar[] = {
	{ 0, 0,				0				}
};

static KEY_BINDING tab_locate[] = {
	{ 0,  CH_CTRL('M'),	cx_locate_go	},
	{ 0,  0,			0				}
};

static KEY_BINDING tab_menu[] = {
	{ 0,  CH_CTRL('M'),	cx_menu_pick	},
	{ 0,  0,			0				}
};

static KEY_BINDING tab_panel[] = {
	{ 0,  KEY_UP,		cx_pan_up		},
	{ 0,  CH_CTRL('A'),	cx_pan_up		},
	{ 0,  KEY_DOWN,		cx_pan_down		},
	{ 0,  CH_CTRL('Z'),	cx_pan_down		},
	{ 0,  KEY_PPAGE,	cx_pan_pgup		},
	{ 0,  KEY_NPAGE,	cx_pan_pgdown	},
	{ 1,  KEY_HOME,		cx_pan_home		},
#ifdef KEY_END
	{ 1,  KEY_END,		cx_pan_end		},
#endif
	{ 0,  0,			0				}
};

static KEY_BINDING tab_select[] = {
	{ 0,  CH_CTRL('M'),	cx_select_files	},
	{ 0,  CH_CTRL('N'),	cx_select_next	},
	{ 0,  CH_CTRL('P'),	cx_select_prev	},
	{ 0,  0,			0				}
};

static KEY_BINDING tab_shortcut[] = {
	/* these lines correspond with the menu panel */
	{ 0,  KEY_F(1),		cx_mode_help		},
	{ 1,  'd',			cx_mode_dir_main	},
	{ 1,  'h',			cx_mode_history		},
	{ 1,  's',			cx_mode_sort		},
	{ 0,  CH_CTRL('F'),	cx_mode_locate		},
	{ 0,  0,			cx_select_allfiles	}, /* no key */
	{ 1,  '+',			cx_mode_select		},
	{ 0,  0,			cx_select_nofiles	}, /* no key */
	{ 1,  '-',			cx_mode_deselect	},
	{ 1,  '*',			cx_select_inverse	},
	{ 1,  '=',			cx_select_compare	},
	{ 1,  'c',			cx_mode_config_main	},
	{ 0,  0,			cx_trans_quit		}, /* no key */
	{ 0,  0,			0					}
};

static KEY_BINDING tab_sort[] = {
	{ 0,  ' ',			cx_sort_set		},
	{ 0,  CH_CTRL('M'),	cx_sort_set		},
	{ 0,  0,			0				}
};

typedef struct {
	int mode;
	void (*prepare_fn)(void);
	KEY_BINDING *table[6];	/* up to 6 tables terminated with NULL;
							   order is sometimes important, only
							   the first KEY_BINDING for a given key
							   is accepted and followed */
} MODE_DEFINITION;

static MODE_DEFINITION mode_definition[] = {
	{ MODE_COMPLETION, compl_prepare,
		{ tab_compl,tab_panel,tab_edit,tab_common,0 } },
	{ MODE_CONFIG_ADMIN, config_prepare,
		{ tab_config,tab_panel,tab_common,0 } },
	{ MODE_CONFIG_CONFIRM, config_confirm_prepare,
		{ tab_config_confirm,tab_panel,tab_common,0 } },
	{ MODE_CONFIG_EDIT_NUM, config_edit_num_prepare,
		{ tab_config_edit_num,tab_edit,tab_common,0 } },
	{ MODE_CONFIG_EDIT_STR, config_edit_str_prepare,
		{ tab_config_edit_str,tab_edit,tab_common,0 } },
	{ MODE_CONFIG_MAIN, config_prepare,
		{ tab_config,tab_panel,tab_common,0 } },
	{ MODE_DIR_EDIT, dir_edit_prepare,
		{ tab_dir,tab_edit,tab_common,0 } },
	{ MODE_DIR_MAIN, dir_main_prepare,
		{ tab_dir,tab_panel,tab_edit,tab_common,0 } },
	{ MODE_DIR_SPLIT, dir_split_prepare,
		{ tab_dir,tab_panel,tab_edit,tab_common,0 } },
	{ MODE_FILE_MAIN, files_main_prepare,
		{ tab_editcmd,tab_shortcut,tab_panel,tab_edit,tab_common,0 } },
	{ MODE_HELP, help_prepare,
		{ tab_help,tab_panel,tab_common,0 } },
	{ MODE_HIST, hist_prepare,
		{ tab_hist,tab_panel,tab_edit,tab_common,0 } },
	{ MODE_LOCATE, locate_prepare,
		{ tab_locate,tab_edit,tab_panel,tab_common,0 } },
	{ MODE_MENU, menu_prepare,
		{ tab_shortcut,tab_menu,tab_panel,tab_common,0 } },
	{ MODE_SELECT, select_prepare,
		{ tab_select,tab_panel,tab_edit,tab_common,0 } },
	{ MODE_SORT, sort_prepare,
		{ tab_sort,tab_panel,tab_common,0 } },
	{ MODE_DESELECT, select_prepare,
		{ tab_select,tab_panel,tab_edit,tab_common,0 } },
	{ 0, 0, { 0 } }
};

/* linked list of all control loop instances */
static struct operation_mode {
	int mode;
	PANEL *panel;
	TEXTLINE *textline;
	struct operation_mode *previous;
} mode_init = { 0,0,0,0 }, *clex_mode = &mode_init;

int
get_current_mode(void)
{
	return clex_mode->mode;
}

int
get_previous_mode(void)
{
	return clex_mode->previous->mode;
}

static MODE_DEFINITION *
get_modedef(int mode)
{
	MODE_DEFINITION *p;

	for (p = mode_definition; p->mode; p++)
		if (p->mode == mode)
			return p;

	err_exit("BUG: requested operation mode %d is invalid",mode);

	/* NOTREACHED */
	return 0;
}

static KEY_BINDING *
action(int ch, KEY_BINDING **tables)
{
	int lower_ch, i, esc, noesc_pos;
	KEY_BINDING **ptab, *tab, *noesc_tab;

	if (ch == CH_ESC)
		return 0;
	esc = kbd_prev() == CH_ESC;
	lower_ch = tolower(ch);
	noesc_tab = 0;
	for (ptab = tables; (tab = *ptab); ptab++) {
		for (i = 0; tab[i].fn; i++)
			if (lower_ch == tab[i].key) {
				if (esc && !tab[i].escp) {
					/*
					 * an entry with 'escp' flag has higher priority,
					 * so we must continue to search the tables to see
					 * if such entry for the given key exists in the
					 * tables.
					 */
				   	if (noesc_tab == 0) {
						/* accept the first definition only */
						noesc_tab = tab;
						noesc_pos = i;
					}
				}
				else if (esc || !tab[i].escp) {
					(*tab[i].fn)();
					return tab;
				}
			}
	}
	if (noesc_tab) {
		(*noesc_tab[noesc_pos].fn)();
		return noesc_tab;
	}

	if (textline && ch > 0 && ch <= 0xFF && !iscntrl(ch)) {
		edit_insertchar(ch);
		return tab_insertchar;
	}

	win_remark("pressed key has no function "
	  "(reminder: F1 - help, ctrl-C - cancel)");
	return 0;
}

/*
 * main control loop for a selected mode 'mode'
 * control loops for different modes are nested whenever necessary
 */
void
control_loop(int mode)
{
	MODE_DEFINITION *modedef;
	KEY_BINDING *tab;
	struct operation_mode current_mode;

	current_mode.previous = clex_mode;
	clex_mode = &current_mode;

	/* panel and textline inherited from previous mode */
	clex_mode->panel = clex_mode->previous->panel;
	clex_mode->textline = clex_mode->previous->textline;

	for (next_mode = mode; /* until break */; ) {
		modedef = get_modedef(clex_mode->mode = next_mode);
		next_mode = 0;
		(*modedef->prepare_fn)();
		win_heading();
		if (panel != clex_mode->panel) {
			clex_mode->panel = panel;
			win_panel();
		}
		if (textline != clex_mode->textline) {
			clex_mode->textline = textline;
			win_edit();
		}

		for (; /* until break */;) {
			tab = action(kbd_input(),modedef->table);
			if (next_mode) {
				if (next_mode == MODE_SPECIAL_RETURN &&
				  clex_mode->previous->mode == 0) {
					win_remark("to quit CLEX press <esc> Q");
					next_mode = 0;
				}
				else
					break;
			}

			/* some special handling not implemented in tables */
			switch (clex_mode->mode) {
			case MODE_MENU:
				if (tab == tab_shortcut)
					next_mode = MODE_SPECIAL_RETURN;
				break;
			case MODE_COMPLETION:
				if (tab == tab_edit || tab == tab_insertchar)
					next_mode = MODE_SPECIAL_RETURN;
				break;
			case MODE_DIR_MAIN:
			case MODE_DIR_SPLIT:
				if (textline->curs)
					next_mode = MODE_DIR_EDIT;
				break;
			case MODE_LOCATE:
				locate_update();
				break;
			}
			if (next_mode)
				break;
		}

		if (next_mode == MODE_SPECIAL_QUIT)
			err_exit("Terminated normally");

		if (next_mode == MODE_SPECIAL_RETURN) {
			next_mode = 0;
			break;
		}

	}

	clex_mode = clex_mode->previous;
	win_heading();
	if (panel != clex_mode->panel) {
		panel = clex_mode->panel;
		pan_adjust(panel);		/* screen size might be resized */
		win_panel();
	}
	if (textline != clex_mode->textline) {
		textline = clex_mode->textline;
		edit_adjust();
		win_edit();
	}
}

void
menu_initialize(void)
{
	/* lines must correspond with tab_shortcut[] */
	static const char *menu[] = {
		"on-line help                             <F1> ",
		"change directory                         <esc> D",
		"command line history                     <esc> H",
		"set file sort order                      <esc> S",
		"locate file                              ctrl-F",
		"select files:  select all                no shortcut",
		"               select using pattern      <esc> +",
		"               deselect all              no shortcut",
		"               deselect using pattern    <esc> -",
		"               inverse selection         <esc> *",
		"               compare directories       <esc> =",
		"configure CLEX                           <esc> C",
		"quit CLEX                                <esc> Q or <esc> X"
	};

	panel_menu.type = PANEL_TYPE_MENU;
	panel_menu.min = -1;
	panel_menu.cnt = ARRAY_SIZE(menu);
	panel_menu.menu = menu;
}

void
menu_prepare(void)
{
	panel_menu.top = panel_menu.curs = -1;
	panel = (PANEL *)&panel_menu;
	textline = 0;
}

void
cx_menu_pick(void)
{
	if (panel_menu.curs >= 0)
		(*tab_shortcut[panel_menu.curs].fn)();
	if (next_mode == 0)
		next_mode = MODE_SPECIAL_RETURN;
}

void 
cx_menu_na(void)
{
	win_remark("ctrl-G works only in the file panel");
}
