/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 1997-2000 Tempest s.r.o. http://www.tempest.sk
 * Copyright (C) 2001-2002 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

#include <autoconf.h>

#include <sys/types.h>		/* time_t */
#include <sys/stat.h>		/* stat() */
#include <errno.h>			/* errno */
#include <locale.h>			/* localeconv() */
#include <stdio.h>			/* sprintf() */
#include <string.h>			/* strcmp() */
#include <unistd.h>			/* stat() */

/* struct tm, time() */
#ifdef TM_IN_SYS_TIME
# include <sys/time.h>
# ifdef TIME_WITH_SYS_TIME
#  include <time.h>
# endif
#else
# include <time.h>
#endif

/* readdir() */
#ifdef HAVE_DIRENT_H
# include <dirent.h>
#else
# define dirent direct
# ifdef HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# ifdef HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# ifdef HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif

/* major() */
#ifdef MAJOR_IN_MKDEV
# include <sys/mkdev.h>
#endif
#ifdef MAJOR_IN_SYSMACROS
# include <sys/sysmacros.h>
#endif

#ifndef S_ISLNK
# define S_ISLNK(X)	(0)
#endif

#include "clex.h"
#include "list.h"

#include "cfg.h"			/* config_num() */
#include "directory.h"		/* filepos_save() */
#include "inout.h"			/* win_warning() */
#include "sort.h"			/* sort_files() */
#include "userdata.h"		/* get_login() */
#include "ustring.h"		/* us_copy() */
#include "util.h"			/* emalloc() */

/*
 * additional FILE_ENTRIES to be allocated when the file panel is full,
 * i.e. when all existing 'fe_slots' in 'files' are occupied
 */
#define FE_ALLOC	128

extern int errno;

static time_t now;
static int alt_a, alt_d, alt_i, alt_l, alt_m;
static int do_a, do_d, do_i, do_l, do_m, do_o, do_p, do_s;
static int clock24, hide, use_pathname = 0;
static char sep000;
static dev_t dirdev;

void
list_initialize(void)
{
	const char *fields;
	char ch, strclock[32];
	time_t h18;

	hide = !config_num(CFG_SHOW_HIDDEN);

	setlocale(LC_ALL,config_str(CFG_LOCALE));
	sep000 = localeconv()->thousands_sep[0];
	if (sep000 != ',')
		sep000 = '.';

	/* 12 or 24 hours clock ? */
	h18 = 64800L;
	strftime(strclock,sizeof(strclock),"%X",gmtime(&h18));
	/* it is either 18:00:00 or 06:00:00 PM */
	clock24 = strchr(strclock,'6') == 0;

	/* which fields are going to be displayed ? */
	alt_a = alt_d = alt_i = alt_l = alt_m = 0;
	do_a = do_d = do_i = do_l = do_m = do_o = do_p = do_s = 0;
	for (fields = config_str(CFG_LAYOUT); (ch = *fields++); )
		switch (ch) {
		case 'A': alt_a = 1;
		case 'a':  do_a = 1; break;
		case 'D': alt_d = 1;
		case 'd':  do_d = 1; break;
		case 'I': alt_i = 1;
		case 'i':  do_i = 1; break;
		case 'L': alt_l = 1;
		case 'l':  do_l = 1; break;
		case 'M': alt_m = 1;
		case 'm':  do_m = 1; break;
		case 'o':  do_o = 1; break;
		case 'p':  do_p = 1; break;
		case 's':  do_s = 1; break;
		/* case 't':  always inspect the file type */
		/* default:   ignore unknown formatting character */
		}
}

/*
 * when calling one of stat2xxx() functions always make sure
 * the string provided as 'str' argument can hold the result.
 * See FE_XXX_STR #defines in clex.h for proper sizes.
 */

static void
stat2time(char *str, time_t tm, int right_align)
{
	struct tm *ptm;
	static char *month[] = {
		"jan", "feb", "mar", "apr", "may", "jun",
		"jul", "aug", "sep", "oct", "nov", "dec"
	};

	ptm = localtime(&tm);
	/* print time of day, after 18 hrs (64800s) print date */
	/* 5 mins (300s) tolerance */
	if (tm <= now + 300 && now <= tm + 64800L)
		if (clock24)
			sprintf(str,right_align ? "  %02d:%02d" : "%02d:%02d  ",
			  ptm->tm_hour,ptm->tm_min);
		else
			sprintf(str,right_align ? " %02d:%02d%c" : "%02d:%02d%c ",
			  1 + (ptm->tm_hour + 11) % 12,ptm->tm_min,
			  ptm->tm_hour < 12 ? 'a' : 'p' );
	else 
		sprintf(str,"%02d%s%02d",ptm->tm_mday,
		  month[ptm->tm_mon],ptm->tm_year % 100);
}

static void
stat2size(char *str, off_t size)
{
	int exp, roundup;

	for (exp = roundup = 0; size > 999999L; exp++) {
		/* size = size / 1024 (1K) */
		size /= 512;
		roundup = size % 2;
		size /= 2;
	}

	sprintf(str," %6ld%c",(long int)(size + roundup)," KMGTE"[exp]);
	/* insert thousands separator */
	if (str[3] != ' ') {
		str[0] = str[1];
		str[1] = str[2];
		str[2] = str[3];
		str[3] = sep000;
	}
}

static void
stat2dev(char *str, int dev_minor, int dev_major)
{
	sprintf(str,DEV_FORMAT,dev_minor,dev_major);
}

static int
stat2type(mode_t mode, uid_t uid)
{
	if (S_ISREG(mode)) {
		if ((mode & S_IXUSR) != S_IXUSR &&
		    (mode & S_IXGRP) != S_IXGRP &&
		    (mode & S_IXOTH) != S_IXOTH)
			return FT_PLAIN_FILE;
		if ((mode & S_ISUID) == S_ISUID)
			return uid ? FT_PLAIN_SUID : FT_PLAIN_SUID_ROOT;
		if ((mode & S_ISGID) == S_ISGID)
			return FT_PLAIN_SGID;
		return FT_PLAIN_EXEC;
	}
	if (S_ISDIR(mode))
		return FT_DIRECTORY;
	if (S_ISBLK(mode))
		return FT_DEV_BLOCK;
	if (S_ISCHR(mode))
		return FT_DEV_CHAR;
	if (S_ISFIFO(mode))
		return FT_FIFO;
#ifdef S_ISSOCK
	if (S_ISSOCK(mode))
		return FT_SOCKET;
#endif
	return FT_OTHER;
}

static void
stat2mode(char *str, mode_t mode)
{
	int mode12;		/* low 12 bits */

	mode12 = mode & 07777;
	/* test: 0755-dir or 0755,0644-file */
	if (alt_m && (mode12 == 0644 && S_ISREG(mode) ||
	  mode12 == 0755 && (S_ISREG(mode) || S_ISDIR(mode))) )
		strcpy(str,"    ");
	else
		sprintf(str,"%04o",mode12);
}

static void
stat2perms(char *str, mode_t mode)
{
	int i;

	strcpy(str,"rwxrwxrwx");
	for (i = 0; i < 9; i++)
		if (!(mode & 1 << i))
			str[8 - i] = '-';
	if ((mode & S_ISUID) == S_ISUID)
		str[2] = (mode & S_IXUSR) == S_IXUSR ? 's' : 'S';
	if ((mode & S_ISGID) == S_ISGID)
		str[5] = (mode & S_IXGRP) == S_IXGRP ? 's' : 'S';
#ifdef S_ISVTX
	if ((mode & S_ISVTX) == S_ISVTX)
		str[8] = (mode & S_IXOTH) == S_IXOTH ? 't' : 'T';
#endif
}

static void
stat2owner(char *str, uid_t uid, gid_t gid)
{
	char number1[16], number2[16];
	const char *pch1, *pch2;

	if ( (pch1 = get_login(uid)) == 0) {
		sprintf(number1,"%u",(unsigned)uid);
		pch1 = number1;
	}
	if ( (pch2 = get_group(gid)) == 0) {
		sprintf(number2,"%u",(unsigned)gid);
		pch2 = number2;
	}
	sprintf(str,"%8.8s:%-8.8s",pch1,pch2);
}

static void
stat2links(char *str, nlink_t nlink)
{
	if (nlink == 1 && alt_l)
		strcpy(str,"    ");
	else if (nlink <= 999)
		sprintf(str,alt_l ? "=%-3d" : " %3d",(int)nlink);
	else
		strcpy(str,">999");
}

static void
space_string(char *str, size_t len)
{
	size_t i;

	for (i = 0; i < len - 1; i++)
		str[i] = ' ';
	str[i] = '\0';
}

/* get the extension "ext" from "file.ext" */
static const char *
get_ext(const char *filename)
{
	const char *ext;
	char ch;

	for (ext = ""; (ch = *filename); filename++)
		if (ch == '.')
			ext = filename + 1;
	return ext;
}

/* this file does exist, but no other information is available */
static void
nofileinfo(FILE_ENTRY *pfe)
{
	pfe->mtime = 0;
	pfe->size = 0;
	pfe->extension = get_ext(USTR(pfe->file));
	pfe->file_type = FT_NA;
	if (do_s)
		space_string(pfe->sd_str.size,sizeof(pfe->sd_str.size));
	if (do_a)
		space_string(pfe->atime_str,sizeof(pfe->atime_str));
	if (do_d)
		space_string(pfe->mtime_str,sizeof(pfe->mtime_str));
	if (do_i)
		space_string(pfe->ctime_str,sizeof(pfe->ctime_str));
	if (do_l)
		space_string(pfe->links_str,sizeof(pfe->links_str));
	if (do_m)
		space_string(pfe->mode1_str,sizeof(pfe->mode1_str));
	if (do_o)
		space_string(pfe->owner_str,sizeof(pfe->owner_str));
	if (do_p)
		space_string(pfe->mode2_str,sizeof(pfe->mode2_str));
}

/* fill-in all required information about a file */
static void
fileinfo(FILE_ENTRY *pfe, struct stat *pst)
{
	pfe->mtime = pst->st_mtime;
	pfe->size = pst->st_size;
	pfe->extension = get_ext(USTR(pfe->file));
	pfe->file_type = stat2type(pst->st_mode,pst->st_uid);
	/* special case: active mounting point */
	if (pfe->file_type == FT_DIRECTORY && !pfe->flag.symlink &&
	  pst->st_dev != dirdev)
		pfe->file_type = FT_DIRECTORY_MNT;

	if (IS_FT_DEV(pfe->file_type))
		/* always do stat2dev() - it is used in sort.c and select.c */
		stat2dev(pfe->sd_str.device,
		  major(pst->st_rdev),minor(pst->st_rdev));
	else if (do_s)
		stat2size(pfe->sd_str.size,pst->st_size);
	if (do_a)
		stat2time(pfe->atime_str,pst->st_atime,!alt_a);
	if (do_d)
		stat2time(pfe->mtime_str,pst->st_mtime,!alt_d);
	if (do_i)
		stat2time(pfe->ctime_str,pst->st_ctime,!alt_i);
	if (do_l)
		stat2links(pfe->links_str,pst->st_nlink);
	if (do_m)
		stat2mode(pfe->mode1_str,pst->st_mode);
	if (do_o)
		stat2owner(pfe->owner_str,pst->st_uid,pst->st_gid);
	if (do_p)
		stat2perms(pfe->mode2_str,pst->st_mode);
}

/* build the FILE_ENTRY '*pfe' describing the file named 'name' */
static int
describe_file(const char *name, FILE_ENTRY *pfe)
{
	struct stat stdata;

	if (lstat(name,&stdata) < 0) {
		if (errno == ENOENT)
			return -1;		/* file deleted in the meantime */
		pfe->flag.symlink = 0;
		nofileinfo(pfe);
		return 0;
	}

	if ( (pfe->flag.symlink = S_ISLNK(stdata.st_mode)) ) {
		if (get_link_us(&pfe->link,name) < 0)
			us_copy(&pfe->link,"??");
		/* need stat() instead of lstat() */
		if (stat(name,&stdata) < 0) {
			nofileinfo(pfe);
			return 0;
		}
	}

	fileinfo(pfe,&stdata);
	return 0;
}

/*
 * We abandoned any form of caching and always build the file
 * panel from scratch. No caching algorithm was 100% perfect,
 * there were always few 'pathological' cases.
 */
static void
directory_read(void)
{
	int i, cnt1, cnt2, dotdir;
	DIR *dd;
	FILE_ENTRY *pfe;
	struct stat st;
	struct dirent *direntry;
	char *name;

	name = USTR(ppanel_file->dir);
	if (stat(name,&st) < 0 || (dd = opendir(name)) == 0) {
		win_warning("READDIR: Cannot list the contents "
		  "of the directory.");
		ppanel_file->cnt = 0;
		return;
	}
	dirdev = st.st_dev;

	win_waitmsg();

	/*
	 * step #1: check selected files already listed in the panel
	 * in order not to loose their selection mark
	 */
	cnt1 = 0;
	for (i = 0; cnt1 < ppanel_file->selected; i++) {
		pfe = ppanel_file->files[i];
		if (!pfe->flag.select)
			continue;
		name = USTR(pfe->file);
		if ((hide && !pfe->flag.dotdir && name[0] == '.') ||
		  describe_file(use_pathname ? pathname_join(name) : name,
		  pfe) < 0)
			/* this entry is no more valid */
			ppanel_file->selected--;
		else {
			/* OK, move it to the end of list we have so far, */
			/* moving by swapping pointers: [cnt1] <--> [i] */
			ppanel_file->files[i] = ppanel_file->files[cnt1];
			ppanel_file->files[cnt1] = pfe;
			cnt1++;
		}
	}

	/* step #2: add data about new files */
	cnt2 = cnt1;
	while ( (direntry = readdir(dd)) ) {
		name = direntry->d_name;
		dotdir = 0;
		if (name[0] == '.') {
			if (name[1] == '\0')
				dotdir = 1;
			else if (name[1] == '.' && name[2] == '\0')
				dotdir = 2;
			else if (hide)
				continue;
		}

		/* didn't we process this file already in step #1 ? */
		if (cnt1) {
			for (i = 0; i < cnt1; i++)
				if (strcmp(USTR(ppanel_file->files[i]->file),name) == 0)
					break;
			if (i < cnt1)
				continue;
		}

		/* allocate new bunch of FILE_ENTRIES when required */
		if (cnt2 == ppanel_file->fe_slots) {
			ppanel_file->fe_slots += FE_ALLOC;
			ppanel_file->files = erealloc(ppanel_file->files,
			  ppanel_file->fe_slots * sizeof(FILE_ENTRY *));
			pfe = emalloc(FE_ALLOC * sizeof(FILE_ENTRY));
			for (i = 0; i < FE_ALLOC; i++) {
				US_INIT(pfe[i].file);
				US_INIT(pfe[i].link);
				ppanel_file->files[cnt2 + i] = pfe + i;
			}
		}

		pfe = ppanel_file->files[cnt2];
		us_copy(&pfe->file,name);
		if (describe_file(use_pathname ? pathname_join(name) : name,
		  pfe) < 0)
			continue;
		pfe->flag.select = 0;
		pfe->flag.dotdir = dotdir;
		cnt2++;
	}
	ppanel_file->cnt = cnt2;

	closedir(dd);
}

void
list_directory(void)
{
	now = time(0);

	/* password data change invalidates data in both panels */
	if (userdata_initialize())
		ppanel_file->other->expired = 1;

	if (ppanel_file->cnt == 0)
		ppanel_file->selected = 0;

	filepos_save();
	directory_read();
	sort_files();
	filepos_set();

	ppanel_file->expired = 0;
}

void
list_alt_directory(void)
{
	/*
	 * warning: primary panel does not correspond with the current
	 * working directory during list_alt_directory(), this is
	 * the only exception
	 */
	panel = (PANEL *)(ppanel_file = ppanel_file->other);
	use_pathname = 1;	/* must prepend directory name */
	pathname_set_directory(USTR(ppanel_file->dir));
	list_directory();
	use_pathname = 0;
	panel = (PANEL *)(ppanel_file = ppanel_file->other);

	/*
	 * note: win_warning() during list_directory() writes
	 * invalid data (from the wrong panel) into the info-line
	 */
}
